from flask import Flask, escape, request, jsonify
from flask_cors import CORS
import os
import webbrowser
from airduct.database import setup_config, fetch_schedules, fetch_flows, fetch_flow, fetch_tasks

def create_app():
    app = Flask(__name__)
    def run_on_start(*args, **argv):
        url = "http://127.0.0.1:5000"
        webbrowser.open(url,new=2)
        print('Opening Browser...')
    # run_on_start()
    return app
app = create_app()
CORS(app, resources={r"/*": {"origins": "*"}})

def run(config):
    os.environ['AIRDUCT_CONFIG_FILE'] = config or ''
    setup_config()
    app.run()

@app.route('/api/schedules')
def schedules():
    schedules = fetch_schedules()
    def transform(schedule):
        return {
            'id': schedule.id,
            'run_at': schedule.run_at,
            'name': schedule.name,
            'originated_file': schedule.originated_file,
            'created_at': None if schedule.created_at is None else schedule.created_at.strftime("%b %d %Y %H:%M:%S"),
            'updated_at': None if schedule.updated_at is None else schedule.updated_at.strftime("%b %d %Y %H:%M:%S")
        }
    return jsonify([transform(x) for x in schedules])

@app.route('/api/flows')
def flows():
    flows = fetch_flows()
    def transform(flow):
        return {
            'id': flow.id,
            'schedule_id': flow.schedule_id,
            'name': flow.name,
            'status': flow.status,
            'created_at': None if flow.created_at is None else flow.created_at.strftime("%b %d %Y %H:%M:%S"),
            'updated_at': None if flow.updated_at is None else flow.updated_at.strftime("%b %d %Y %H:%M:%S")
        }
    return jsonify([transform(x) for x in flows])

@app.route('/api/flows/<string:names>')
def get_flows(names):
    flows = fetch_flow(names)
    def transform(flow):
        return {
            'id': flow.id,
            'schedule_id': flow.schedule_id,
            'name': flow.name,
            'status': flow.status,
            'created_at': None if flow.created_at is None else flow.created_at.strftime("%b %d %Y %H:%M:%S"),
            'updated_at': None if flow.updated_at is None else flow.updated_at.strftime("%b %d %Y %H:%M:%S")
        }
    return jsonify([transform(x) for x in flows])


@app.route('/api/tasks/<string:flow_id>')
def tasks(flow_id):
    tasks = fetch_tasks(flow_id)
    def transform(task):
        return {
            'id': task.id,
            'flow_id': task.flow_id,
            'step': task.step,
            'status': task.status,
            'can_fail': task.can_fail,
            'completed_at': None if task.completed_at is None else task.completed_at.strftime("%b %d %Y %H:%M:%S"),
            'created_at': None if task.created_at is None else task.created_at.strftime("%b %d %Y %H:%M:%S"),
            'updated_at': None if task.updated_at is None else task.updated_at.strftime("%b %d %Y %H:%M:%S")
        }
    return jsonify([transform(x) for x in tasks])



from airduct.database import Schedule, Task, Flow, get_session
import inspect
import pickle
import os

class TaskNotFoundError(AttributeError):
    pass


class FlowNestingTooDeep(Exception):
    pass


def schedule(name, run_at, flow):
    session = get_session()
    frame = inspect.stack()[1]
    module = inspect.getmodule(frame[0])
    _schedule = Schedule(name=name, run_at=run_at, flow=pickle.dumps(flow), originated_file=module.__name__)
    for _task in flow:
        if type(_task) is list:
            for t in _task:
                if type(t) is list:
                    raise FlowNestingTooDeep('Reached max nesting level in flow. Only 2 levels allowed')
    r = session.query(Schedule).filter(Schedule.name == name).first()
    if r is not None:
        # There is already a schedule, let's update it
        r.run_at = run_at
        r.flow = pickle.dumps(flow)
        session.commit()
        return
    session.add(_schedule)
    session.commit()


def task(name, can_fail=False):
    frm = inspect.stack()[1]
    mod = inspect.getmodule(frm[0])
    return pickle.dumps({'module': mod.__name__, 'func_name': name, 'can_fail': can_fail})



import importlib
import pkgutil


def import_submodules(package, recursive=True):
    if isinstance(package, str):
        package = importlib.import_module(package)
    results = []
    for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = package.__name__ + '.' + name
        results.append(full_name)
        if recursive and is_pkg:
            results = results + import_submodules(full_name)
    return results


def find_flow_files(folder_path):
    modules = import_submodules(folder_path)
    for module_name in modules:
        importlib.import_module(module_name)
    return modules



import subprocess

import click
import asyncio
from airduct.core import start_scheduler, start_worker, trigger_schedule
from airduct.api.api import run as api_run
loop = asyncio.get_event_loop()


@click.group()
def cli():
    pass


@click.command(help="Start a worker")
@click.option('--config', help='Path to your config.yml')
def worker(config):
    click.echo('Starting a Worker')
    start_worker(config)


@click.command(help="Start the scheduler")
@click.option('-p', '--path', '_path', prompt=True)
@click.option('-c', '--config', '_config', help='Path to your config.yml')
@click.option('-w', '--run_with_worker', '_no_worker', default=False)
def scheduler(_path, _config, _no_worker):
    click.echo('Starting Scheduler...')
    click.echo(f'Loding config from: {_config}')
    start_scheduler(_path, _config, _no_worker)


@click.command(help="Trigger a schedule by name")
@click.option('-c', '--config', '_config', help='Path to your config.yml')
@click.argument('name')
def trigger(_config, name):
    trigger_schedule(_config, name)
    click.echo('Run a schedule by name')


@click.command(help="Start the webapp's api server")
@click.option('-c', '--config', '_config', help='Path to your config.yml')
def api(_config):
    api_run(_config)


@click.command(help="Build a version of the webapp for production")
@click.option('-H', '--host', prompt=True, help="BaseURL for API. default: http://localhost:5000", default="http://localhost:5000")
def webapp(host):
    from pathlib import Path
    click.echo('You must have node/yarn installed')
    cwd = str(Path(__file__).parent.parent.parent) + '/webapp'
    subprocess.call(
        ['yarn', 'install'],
        cwd=cwd,
    )
    subprocess.call(
        ['yarn', 'build'],
        cwd=cwd,
        env={'REACT_APP_API_URL': host}
    )
    subprocess.call(
        ['mv', cwd+'/build', 'build']
    )

    click.echo('To serve now, you can use: $ cd build && python3 -m http.server')


cli.add_command(worker)
cli.add_command(scheduler)
cli.add_command(trigger)
cli.add_command(api)
cli.add_command(webapp)

if __name__ == '__main__':
    cli()




# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['airduct', 'airduct.api', 'airduct.cli']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'crontab>=0.22.6,<0.23.0',
 'flask>=1.1,<2.0',
 'flask_cors>=3.0,<4.0',
 'psycopg2>=2.8,<3.0',
 'pyyaml>=5.1,<6.0',
 'sqlalchemy>=1.3,<2.0']

entry_points = \
{'console_scripts': ['airduct = airduct.cli.run:cli']}

setup_kwargs = {
    'name': 'airduct',
    'version': '0.1.9',
    'description': 'Simple Pipeline Scheduler in Python',
    'long_description': '# airduct\nSimple Pipeline Scheduler in Python\n\n![Airduct Screenshot](docs/screenshot.png)\n\n## Links\n\n- [Github](https://github.com/alairock/airduct)\n- [Documentation](https://airduct.readthedocs.io)\n\n## Installing\n    $ pip install airduct\n\nor\n\n    $ poetry add airduct\n\n## Quickstart\nAirduct calls pipelines "flows". A flow is a python file with a very particular definition, which is hopefully self explanatory.\n\nHere is an example flow:\n\n```python\nfrom airduct import schedule, task\n\n\nschedule(\n    name=\'ExampleFlow\',\n    run_at=\'* * * * *\',\n    flow=[\n        task(\'e1f1\'),\n        [task(\'e1f2\'), task(\'e1f3\', can_fail=True)],\n        [task(\'e1f4\')]\n    ]\n)\n\nasync def e1f1():\n    print(\'e1f1 - An async function!\')\n\ndef e1f2():\n    print(\'e1f2 - Regular functions work too\')\n\nasync def e1f3():\n    print(\'e1f3\')\n\nasync def e1f4():\n    print(\'e1f4\')\n```\n\nA flow requires a `airduct.scheduling.schedule` which runs at scheduler initialization. \nThe schedule function requires:\n - `name`: A name to identify the flow as, must be unique\n - `run_at`: A crontab-like scheduling syntax. (Uses this [crontab parser](https://github.com/josiahcarlson/parse-crontab))\n - `flow`: A list of `airduct.scheduling.task`\'s. These can be nested lists, for parallel tasks, 2 levels deep. See example.\n\n`task()` Requires the name of the function you desire to run during that step. Must be defined in the same flow file. You can ignore errors with `can_fail=True` in the function\'s signature.\n\nThis file is placed in a folder/python-module alongside other flows.\n\nTo run: `$ airduct schedule --path /path/to/flow/folder`\n\nBy default it uses a sqlite in-memory database. If using the in-memory database, it will also automatically run as a worker, in addition to a scheduler.\n\n',
    'author': 'Skyler Lewis',
    'author_email': 'skyler.lewis@canopytax.com',
    'url': 'https://github.com/alairock/airduct',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import pickle
import datetime
from airduct.logger import logger
from airduct.config import getenv
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine, Column, Integer, String, DateTime, Boolean, and_, PickleType
from sqlalchemy.sql import func


Base = declarative_base()
ENGINE = None
Session = sessionmaker()
SeshCache = None


def setup_config():
    global ENGINE, Session
    verbose = getenv('DB_VERBOSE', False)
    encoding = getenv('DB_ENCODING', 'utf-8')
    dialect = '+'.join([x for x in [getenv('DB_DIALECT', 'sqlite'), getenv('DB_DRIVER', None)] if x is not None])
    credentials = ':'.join([getenv('DB_USER', ''), getenv('DB_PASS', '')]) + '@'
    hostname = getenv('DB_HOSTNAME')
    dbport = ':' + str(getenv('DB_PORT')) if getenv('DB_PORT') is not None else ''
    dbname = '/' + getenv('DB_NAME') if getenv('DB_NAME') is not None else ''
    sqlite_file = getenv('DB_SQLITE_FILE', ':memory:')
    colon_style = f':///{sqlite_file}' if getenv('DB_DIALECT', 'sqlite') == 'sqlite' else '://'
    dsn = getenv('DB_DSN', f'{dialect}{colon_style}{credentials}{hostname}{dbport}{dbname}')
    if getenv('DB_DIALECT', 'sqlite') == 'sqlite':
        dsn = 'sqlite' + colon_style
    Session = sessionmaker()
    ENGINE = create_engine(dsn, encoding=encoding, echo=verbose)
    Session.configure(bind=ENGINE)


def get_session():
    global SeshCache
    if SeshCache is None:
        SeshCache = Session()
        return SeshCache
    return SeshCache

class Task(Base):
    __tablename__ = 'tasks'

    id = Column(Integer, primary_key=True)
    status = Column(String)
    flow_id = Column(Integer)
    parents = Column(PickleType)
    step = Column(Integer)
    pickled_task = Column(PickleType)
    can_fail = Column(Boolean)
    logs = Column(String)
    completed_at = Column(DateTime)
    created_at = Column(DateTime, server_default=func.now())
    updated_at = Column(DateTime, onupdate=func.now())


class Flow(Base):
    __tablename__ = 'flows'

    id = Column(Integer, primary_key=True)
    schedule_id = Column(Integer)
    status = Column(String)
    name = Column(String)
    created_at = Column(DateTime, server_default=func.now())
    updated_at = Column(DateTime, onupdate=func.now())


class Schedule(Base):
    __tablename__ = 'schedules'

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)
    run_at = Column(String)
    flow = Column(PickleType)
    originated_file = Column(String)
    created_at = Column(DateTime, server_default=func.now())
    updated_at = Column(DateTime, onupdate=func.now())


def initdb():
    setup_config()
    logger.info('Setting up database tables')
    Base.metadata.create_all(ENGINE)
    logger.info('Complete setting up database tables')
    return get_session()


def fetch_schedules():
    session = get_session()
    results = session.query(Schedule).all()
    return results


def fetch_schedule(name):
    session = get_session()
    results = session.query(Schedule).filter(Schedule.name == name).first()
    return results

def fetch_flows():
    session = get_session()
    current_time = datetime.datetime.utcnow()
    yesterday = current_time - datetime.timedelta(days=7)
    return session.query(Flow) \
        .filter(Flow.created_at >= yesterday) \
        .order_by(Flow.created_at.desc()) \
        .limit(10) \
        .all()

def fetch_flow(name):
    session = get_session()
    return session.query(Flow).filter(Flow.name.in_([name]))\
        .order_by(Flow.created_at.desc()).all()

def fetch_tasks(flow_id):
    session = get_session()
    return session.query(Task).filter(Task.flow_id == flow_id)\
        .order_by(Task.step.asc()).all()

def has_concurrent_flow(schedule):
    session = get_session()
    r = session.query(Flow).filter(Flow.name.in_([schedule.name])).filter(Flow.status.in_(['InProgress'])).all()
    if len(r) > 0:
        return True
    return False


def create_flow_record(schedule):
    session = get_session()
    flow = Flow(name=schedule.name, schedule_id=schedule.id, status='InProgress')
    session.add(flow)
    session.commit()
    return flow


def create_task_record(flow_id, parents, step, task, pickled_task):
    session = get_session()
    task = Task(
        flow_id=flow_id,
        parents=None if len(parents) == 0 else pickle.dumps(parents),
        step=step,
        can_fail=task.get('can_fail', False),
        pickled_task=pickled_task
    )
    session.add(task)
    session.commit()
    return task


def fetch_task():
    session = get_session()
    # TODO: Check if flow is older than timeout period. If it is, then don't include in these results
    results = session.query(Flow, Task) \
        .join(Task, and_(Task.flow_id == Flow.id, Task.status.is_(None)), isouter=True) \
        .filter(Flow.status == 'InProgress') \
        .filter(Task.parents.is_(None)) \
        .order_by(Task.step) \
        .first()

    if results is None:
        return results
    task = results[1]
    if task is None:
        return task
    task.status = 'InProgress'
    session.commit()
    return task


def mark_task(task, status="Complete", can_fail=False, message=''):
    session = get_session()
    session.query(Task).filter(Task.id == task.id).update({
        'status': status,
        'completed_at': datetime.datetime.now(),
        'logs': message
    })
    session.commit()
    if status is 'Error' and not can_fail:
        mark_flow(task.flow_id, status=Error)
        return
    _update_children(task)


def _update_children(task):
    session = get_session()
    results = session.query(Task).filter(Task.flow_id == task.flow_id).filter(Task.step == int(task.step)+1).all()
    for r in results:
        if not isinstance(r.parents, str):
            parents = None if r.parents is None else pickle.loads(r.parents)
            new_parents = None if parents is None else parents.remove(task.id)
            r.parents = None if new_parents is None or len(new_parents) == 0 else pickle.dumps(new_parents)
    session.commit()
    if len(results) == 0:
        logger.info(f'Flow ended: {task.flow_id}')
        mark_flow(task.flow_id)
        return


def mark_flow(flow_id, status='Complete'):
    session = get_session()
    session.query(Flow).filter(Flow.id == flow_id).update({
        'status': status
    })
    session.commit()



import os
import yaml
from airduct import logger

ENV_CONFIG = None


def getenv(name, default=None):
    global ENV_CONFIG
    env_prefix = os.getenv('ENV_PREFIX', '')
    value = os.getenv(env_prefix + name)
    if value is not None:
        return value

    c = os.environ.get('AIRDUCT_CONFIG_FILE')
    config_file_value = ENV_CONFIG
    if ENV_CONFIG is None:
        try:
            with open(c) as file:
                ENV_CONFIG = yaml.load(file, Loader=yaml.FullLoader) or {}
                config_file_value = ENV_CONFIG
        except FileNotFoundError:
            ENV_CONFIG = {}
            config_file_value = {}
            logger.logger.info(
                f'Config file not found {c}, will attempt to use env vars.')

    name_split = name.split('_')
    for index, v in enumerate(name_split):
        config_file_value = config_file_value.get(v.lower(), None if index + 1 == len(name_split) else {})

    if config_file_value is None:
        return default
    return config_file_value



import asyncio
import os
import pickle
import crontab
import importlib
from airduct import logger
from airduct.flowfinder import find_flow_files
from airduct.database import mark_task, has_concurrent_flow, create_flow_record, \
    create_task_record, fetch_task, initdb, fetch_schedules, fetch_schedule, setup_config
import warnings
import time
import datetime
from airduct.logger import logger


warnings.simplefilter(action='ignore', category=FutureWarning)


def start_flow(schedule):
    if has_concurrent_flow(schedule):
        logger.info('Job in progress, skipping to prevent duplicate work.')
        return
        # TODO: Maybe allow multiple flows to run, but this needs to be a configuration

    flow = create_flow_record(schedule)
    step = 0
    parents = []
    logger.info(f'Starting flow: {flow.id}, schedule {flow.schedule_id}')
    for task in pickle.loads(schedule.flow):
        if type(task) is list:
            future_parents = []
            for t in task:
                _task = create_task_record(flow.id, parents, step, pickle.loads(t), t)
                future_parents.append(_task.id)
        else:
            _task = create_task_record(flow.id, parents, step, pickle.loads(task), task)
            future_parents = [_task.id]
        parents = future_parents
        step += 1


def start_scheduler(path, config, no_worker=False):
    os.environ['AIRDUCT_CONFIG_FILE'] = config or ''
    session = initdb()
    flows = find_flow_files(path)
    schedules = fetch_schedules()
    # We use set.difference() here instead of ^ since we only want what is the database and not in the files
    removed_modules = set([x.originated_file for x in schedules]).difference(set(flows))
    while True:
        for schedule in schedules:
            if schedule.originated_file in removed_modules:
                continue
            if crontab.CronTab(schedule.run_at).test(datetime.datetime.now()):
                start_flow(schedule)
        if no_worker or str(session.bind.url) == 'sqlite:///:memory:':
            work_on_tasks()
        time.sleep(1)


def trigger_schedule(config, name):
    os.environ['AIRDUCT_CONFIG_FILE'] = config or ''
    setup_config()
    schedule = fetch_schedule(name)
    if schedule is not None:
        start_flow(schedule)


def start_worker(config):
    os.environ['AIRDUCT_CONFIG_FILE'] = config
    initdb()
    while True:
        if work_on_tasks() == 'sleep':
            time.sleep(1)  # No work to do, let's not cruise so fast, and pound the db.
            # Maybe make this configurable in the futer

def _load_module(task):
    module = None
    try:
        task_definition = pickle.loads(task.pickled_task)
        module = importlib.import_module(task_definition['module'])
        task_func = getattr(module, task_definition['func_name'])
        return task_func
    except AttributeError:
        msg = f'{task.function_name} not found in {module.__file__}'
        raise Exception(msg)
    except ModuleNotFoundError as e:
        raise Exception(f'Module not found, maybe it doesnt exist anymore? '
                        f'{task_definition["module"]}')


def work_on_tasks():
    task = fetch_task()
    if task is None:
        return 'sleep'

    logger.info(f'Starting task: {task.id}')
    try:
        task_func = _load_module(task)
        # TODO: If task is taking too long, then kill it. This would be a configurable field on the Task object.
        # eg: `if start_time + task.timout > now(): Kill task_func
        if asyncio.iscoroutinefunction(task_func):
            loop = asyncio.get_event_loop()
            loop.run_until_complete(task_func())
        else:
            task_func()
    except Exception as e:
        if not task.can_fail:
            logger.error(f'Something bad happened. {e}')
            mark_task(task, 'Error', message=str(e))
            return
        logger.debug(f'An error happened but your task configuration allowed it. {e}')
        mark_task(task, 'Error', message=str(e))
        return
    mark_task(task)
    logger.info(f'Task finished: {task.id}')
